#include <kernel.h>
#include <arch/mm.h>
#include <arch/descriptor.h>
#include <arch/irq.h>
#include <arch/idt.h>

void idle_task(void);

#define load_idt(idtr) \
	asm volatile("lidt (%0)": :"r" (idtr));

/* IDT structure */
dt_entry_t __desc_aligned IDT[256];
struct gdtr loadidt={ sizeof(IDT)-1, (u_int32_t)IDT};

/* Add a new trap vector to the IDT  */
void idt_trap(void *handler, u_int8_t intr)
{
	IDT[intr].gate.selector		= 0x08;
	IDT[intr].gate.offset_low	= (u_int16_t)(((u_int32_t)handler)&0xffff);
	IDT[intr].gate.offset_high 	= (u_int16_t)(((u_int32_t)handler)>>16);
	IDT[intr].gate.access		= D_PRESENT | D_TRAP | D_DPL3;
}

/* Add a new interrupt vector to the IDT  */
void idt_interrupt(void *handler, u_int8_t intr)
{
	IDT[intr].gate.selector		= 0x08;
	IDT[intr].gate.offset_low	= (u_int16_t)(((u_int32_t)handler)&0xffff);
	IDT[intr].gate.offset_high 	= (u_int16_t)(((u_int32_t)handler)>>16);
	IDT[intr].gate.access		= D_PRESENT | D_INT;
}

void int_null(void);
asm (
	".globl int_null\n"
	"int_null:\n"
	"iret\n"
);

void idt_init(void)
{
	int count;

	/* Stick all our vectors in to the IDT */
	idt_trap(_exc00, 0x00);
	idt_trap(_exc01, 0x01);
	idt_trap(_exc02, 0x02);
	idt_trap(_exc03, 0x03);
	idt_trap(_exc04, 0x04);
	idt_trap(_exc05, 0x05);
	idt_trap(_exc06, 0x06);
	idt_trap(_exc07, 0x07);
	idt_trap(_exc08, 0x08);
	idt_trap(_exc09, 0x09);
	idt_trap(_exc0a, 0x0a);
	idt_trap(_exc0b, 0x0b);
	idt_trap(_exc0c, 0x0c);
	idt_trap(_exc0d, 0x0d);
	idt_trap(_exc0e, 0x0e);
	idt_trap(_exc0f, 0x0f);
	idt_trap(_exc10, 0x10);
	idt_trap(_exc11, 0x11);
	idt_trap(_exc12, 0x12);
	idt_trap(_exc13, 0x13);

	for(count=0x14; count<256; count++)
		idt_trap(int_null, count);

	/* IRQs from the 8259A PIC */
	idt_interrupt(_irq0, 0x20);
	idt_interrupt(_irq1, 0x21);
	idt_interrupt(_irq3, 0x23);
	idt_interrupt(_irq4, 0x24);
	idt_interrupt(_irq5, 0x25);
	idt_interrupt(_irq6, 0x26);
	idt_interrupt(_irq7, 0x27);
	idt_interrupt(_irq8, 0x28);
	idt_interrupt(_irq9, 0x29);
	idt_interrupt(_irq10, 0x3a);
	idt_interrupt(_irq11, 0x2b);
	idt_interrupt(_irq12, 0x2c);
	idt_interrupt(_irq13, 0x2d);
	idt_interrupt(_irq14, 0x2e);
	idt_interrupt(_irq15, 0x2f);

	load_idt(&loadidt);
}

void _exc00()
{
	printk("(int:0x00): Divide Error\n");
	idle_task();
}

void _exc01()
{
	printk("(int:0x01): Debug exception\n");
	idle_task();
}

void _exc02()
{
	printk("(int:0x02): Non-Maskable Interrupt\n");
	idle_task();
}

void _exc03()
{
	printk("(int:0x03): Breakpoint\n");
	idle_task();
}

void _exc04()
{
	printk("(int:0x04): Overflow\n");
	idle_task();
}

void _exc05()
{
	printk("(int:0x05): Bounds check\n");
	idle_task();
}

void _exc06()
{
	printk("(int:0x06): Invalid opcode\n");
	idle_task();
}

void _exc07()
{
	printk("(int:0x07): Coprocessor not avaulable\n");
	idle_task();
}

void _exc08()
{
	printk("(int:0x08): Double fault\n");
	idle_task();
}

void _exc09()
{
	printk("(int:0x09): Coprocessor segment overrun\n");
	idle_task();
}

void _exc0a()
{
	printk("(int:0x0a): Invalid TSS\n");
	idle_task();
}

void _exc0b()
{
	printk("(int:0x0b) Segment not present\n");
	idle_task();
}

void _exc0c()
{
	printk("(int0x0c): Stack exception\n");
	idle_task();
}

void _exc0d()
{
	printk("(int:0x0d): General Protection Fault (w00p!)\n");
	idle_task();
}

void _exc0e()
{
	printk("(int:0x0e): Page Fault");
	idle_task();
}

void _exc0f()
{
	printk("(int:0x0f): Unknown error\n");
	idle_task();
}

void _exc10()
{
	printk("(int:0x10): Coprocessor error\n");
	idle_task();
}

void _exc11()
{
	printk("(int:0x11): Alignment check\n");
	idle_task();
}

void _exc12()
{
	printk("(int:0x12): Machine check\n");
	idle_task();
}

void _exc13()
{
	printk("(int:0x13): SSE fault\n");
	idle_task();
}
